import numpy as np


class TireForces(object):
    def __init__(self, fx =np.zeros(4), fy =np.zeros(4), wheel_forces_transformed_force2vehicle_sys = np.zeros((3, 4), dtype=float)):        
        # Dynamic forces on the tires
        # pag 272 eq.54
        self.fx = fx
        self.fy = fy
        self.wheel_forces_transformed_force2vehicle_sys = wheel_forces_transformed_force2vehicle_sys 

    def __array__(self) -> np.ndarray:
        return np.array([*self.fx,
                         *self.fy,
                         *self.wheel_forces_transformed_force2vehicle_sys.flatten()])



import numpy as np


class StateVector(object):
    def __init__(self, x=0., y=0., z=0., roll=0., pitch=0., yaw=0., vx=0., vy=0., vz=0., wx=0., wy=0., wz=0., acc_x = 0, acc_y=0, acc_z=0):

        # Position and Euler angles
        self._x = x
        self._y = y 
        self._z = z 
        self.roll = roll
        self.pitch = pitch
        self.yaw = yaw
        self.vx = vx
        self.vy = vy
        self.vz = vz
        self.wx = wx 
        self.wy = wy 
        self.wz = wz
        self.acc_x = acc_x
        self.acc_y = acc_y
        self.acc_z = acc_z
        self.wx_dot = 0
        self.wy_dot = 0
        self.wz_dot = 0

    def __array__(self) -> np.ndarray:
        return np.array([self._x,
                         self._y,
                         self._z,
                         self.roll,
                         self.pitch,
                         self.yaw,
                         self.vx,
                         self.vy,
                         self.vz,
                         self.wx,
                         self.wy,
                         self.wz,
                         self.acc_x,
                         self.acc_y,
                         self.acc_z,
                         self.wx_dot,
                         self.wy_dot,
                         self.wz_dot])

    def __repr__(self):
        return f"StateVector with x {self.x} y {self.y} z {self.z} roll {self.roll} pitch {self.pitch} yaw {self.yaw} vx {self.vx} vy {self.vy} vz {self.vz} wx {self.wx} wy {self.wy} wz {self.wz} acc_x {self.acc_x} acc_y {self.acc_y} acc_z {self.acc_z} wx_dot {self.wx_dot} wy_dot {self.wy_dot} wz_dot {self.wz_dot}"
        pass

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @x.setter
    def x(self, value):
        if np.isnan(value).any():
            raise ValueError("x_a was set as NaN")
        self._x = value    

    @y.setter
    def y(self, value):
        if np.isnan(value).any():
            raise ValueError("x_a was set as NaN")
        self._y = value

    @z.setter
    def z(self, value):
        if np.isnan(value).any():
            raise ValueError("x_a was set as NaN")
        self._z = value

    @property
    def roll(self):
        return self._roll

    @property
    def pitch(self):
        return self._pitch

    @property
    def yaw(self):
        return self._yaw

    @roll.setter
    def roll(self, value):
        if np.isnan(value).any():
            raise ValueError("x_a was set as NaN")
        self._roll = value
        self.sin_roll = np.sin(value)
        self.cos_roll = np.cos(value)

    @pitch.setter
    def pitch(self, value):
        if np.isnan(value).any():
            raise ValueError("x_a was set as NaN")
        self._pitch = value
        self.sin_pitch = np.sin(value)
        self.cos_pitch = np.cos(value)

    @yaw.setter
    def yaw(self, value):
        if np.isnan(value).any():
            raise ValueError("x_a was set as NaN")
        self._yaw = value
        self.sin_yaw = np.sin(value)
        self.cos_yaw = np.cos(value)



import numpy as np


class StrutForce(object):
    def __init__(self, f_za=np.zeros(4), f_za_dot=np.zeros(4), spring_force = np.zeros(4), dumper_force = np.zeros(4)):  # TODO calculate forces on the strut
        self.f_za = f_za
        self.f_za_dot = f_za_dot
        self.dumper_force = dumper_force 
        self.spring_force = spring_force 

    def __array__(self) -> np.ndarray:
        return np.array([*self.f_za.flatten(),
                         *self.f_za_dot.flatten(),
                         *self.dumper_force.flatten(),
                         *self.spring_force.flatten()])



from vehicle_dynamics.utils.StaticParameters import StaticParameters
from vehicle_dynamics.structures.StateVector import StateVector
from vehicle_dynamics.utils.CurrentStates import CurrentStates
from vehicle_dynamics.utils.import_data_CM import import_data_CM
from vehicle_dynamics.utils.LocalLogger import LocalLogger
from vehicle_dynamics.utils.plot_function import plot_function
from copy import copy
import matplotlib.pyplot as plt
import numpy as np
import logging
import yaml 


class Wheels:
    def __init__(self, static_parameters, logger: logging.Logger):
        self.static_parameters = static_parameters
        # x-position from Vehicle CoG to the front axle [m]
        self.long_f = static_parameters.body.lf
        # x-position from Vehicle Cog to the rear axle [m]
        self.long_r = - static_parameters.body.lr
        # y-position from Vehicle CoG to the left chassis point [m]
        self.lat_l = static_parameters.body.wl
        # y-position from Vehicle CoG to the right chassis point  [m]
        self.lat_r = -static_parameters.body.wr
        pass

    def wheels(self, current_state: CurrentStates, steering_input: float):
        """
        wheels is a function that calculates the current angle
        of the frontal wheel based on the vehicle coordinate system,
        where the output delta is steering angle of the frontal wheels.

        This method calculate the wheel slip on each wheel. This is done using the relative 
        velocity between the wheel angular speed and the vehicle speed.

        Tire model calculates the wheel forces fx and fy using the Magic Formula

        Required current_state from static_parameters:
            1. steering_lock
            2. steering_ratio
            3. dynamic_radius
        Required Arguments:
            1.steering_inputtire.
            2. x_a
                2.01 vx
                2.02 vy
            3. wheel_w_vel
        Returns:
            1. delta
            2. slip_x
            3. slip_y

        """
        # Convert Steering input [-1,1] to wheel steering (delta)
        steering_angle = steering_input * self.static_parameters.steering.maximum_steering_angle 
        current_state.delta = steering_angle / self.static_parameters.steering.ratio

        # Slip Calculation
        MINIMUM_SPEED_VALUE = 15
        if (abs(self.static_parameters.tire.dynamic_radius * current_state.wheel_w_vel).all() <= MINIMUM_SPEED_VALUE) and (abs(current_state.x_a.vx) <= MINIMUM_SPEED_VALUE):
            current_state.slip_x = ((((self.static_parameters.tire.dynamic_radius * current_state.wheel_w_vel) - current_state.x_a.vx) / (MINIMUM_SPEED_VALUE)))
        else:
            # equation 11.30 Bardini
            current_state.slip_x = ((((self.static_parameters.tire.dynamic_radius * current_state.wheel_w_vel) - current_state.x_a.vx) / np.maximum(np.absolute(self.static_parameters.tire.dynamic_radius * current_state.wheel_w_vel), np.absolute(current_state.x_a.vx))))


        previous_slip_y = copy(current_state.slip_y)
        # Refenrence from VTI Y Slip angle
        if (current_state.x_a.vx <= MINIMUM_SPEED_VALUE and current_state.x_a.wz * self.lat_r <= MINIMUM_SPEED_VALUE):
            current_state.slip_y[0] = (current_state.delta - np.arctan((current_state.x_a.vy + self.long_f * current_state.x_a.wz) / MINIMUM_SPEED_VALUE))   # Front Left
            current_state.slip_y[1] = (- np.arctan((current_state.x_a.vy + self.long_r * current_state.x_a.wz) / MINIMUM_SPEED_VALUE))  # Rear Left
            current_state.slip_y[2] = (current_state.delta - np.arctan((current_state.x_a.vy + self.long_f * current_state.x_a.wz) / MINIMUM_SPEED_VALUE)) # Front Right
            current_state.slip_y[3] = (- np.arctan((current_state.x_a.vy + self.long_r * current_state.x_a.wz) / MINIMUM_SPEED_VALUE))  # Rear Right
        else:
            current_state.slip_y[0] = (current_state.delta - np.arctan((current_state.x_a.vy + self.long_f * current_state.x_a.wz) / (current_state.x_a.vx - self.lat_l * current_state.x_a.wz))) #- self.static_parameters.tire.relaxation*current_state.slip_y_rate[0]/current_state.x_a.vx   # Front Left
            current_state.slip_y[1] = (- np.arctan((current_state.x_a.vy + self.long_r * current_state.x_a.wz) / (current_state.x_a.vx - self.lat_l * current_state.x_a.wz))) #- self.static_parameters.tire.relaxation*current_state.slip_y_rate[1]/current_state.x_a.vx  # Rear Left
            current_state.slip_y[2] = (current_state.delta - np.arctan((current_state.x_a.vy + self.long_f * current_state.x_a.wz) / (current_state.x_a.vx - self.lat_r * current_state.x_a.wz))) #- self.static_parameters.tire.relaxation*current_state.slip_y_rate[2]/current_state.x_a.vx   # Front Right
            current_state.slip_y[3] = (- np.arctan((current_state.x_a.vy + self.long_r * current_state.x_a.wz) / (current_state.x_a.vx - self.lat_r * current_state.x_a.wz))) #- self.static_parameters.tire.relaxation*current_state.slip_y_rate[3]/current_state.x_a.vx # Rear Right

        current_state.slip_y_rate = (previous_slip_y - current_state.slip_y)/self.static_parameters.time_step

        # Tire Model

        current_state.x_rf.fx = current_state.f_zr.wheel_load_z * self.static_parameters.tire.longitudinal.peak_friction * np.sin(self.static_parameters.tire.longitudinal.shape_factor * np.arctan((self.static_parameters.tire.longitudinal.slip_stiffness / (self.static_parameters.tire.longitudinal.shape_factor * self.static_parameters.tire.longitudinal.peak_friction)) * current_state.slip_x))
        current_state.x_rf.fy = current_state.f_zr.wheel_load_z * self.static_parameters.tire.lateral.peak_friction * np.sin(self.static_parameters.tire.lateral.shape_factor * np.arctan((self.static_parameters.tire.lateral.cornering_coefficient / (self.static_parameters.tire.lateral.shape_factor * self.static_parameters.tire.lateral.peak_friction)) * current_state.slip_y))



        current_state.compiled_wheel_forces = np.array([current_state.x_rf.fx, current_state.x_rf.fy, current_state.f_zr.wheel_load_z])

        delta = np.array([current_state.delta, 0.0, current_state.delta, 0.0])  # FL, RL, FR, RR
        for i in range(4):
            current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[0, i] = current_state.x_rf.fx[i] * np.cos(delta[i]) - current_state.x_rf.fy[i] * np.sin(delta[i])
            current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[1, i] = current_state.x_rf.fy[i] * np.cos(delta[i]) + current_state.x_rf.fx[i] * np.sin(delta[i]) 

        final_ratio = self.static_parameters.powertrain.gearbox.gear_ratio[current_state.gear] * self.static_parameters.powertrain.differential.ratio
        current_state.x_rr.pho_r_2dot = (current_state.powertrain_net_torque - current_state.x_rf.fx * self.static_parameters.tire.dynamic_radius - self.static_parameters.tire.rolling_resistance_coefficient*current_state.f_zr.wheel_load_z) / (self.static_parameters.tire.inertia + self.static_parameters.powertrain.gearbox.inertia * final_ratio ** 2 + self.static_parameters.powertrain.differential.driveshaft_inertia * self.static_parameters.powertrain.differential.ratio ** 2)
        current_state.wheel_w_vel = current_state.wheel_w_vel + (current_state.x_rr.pho_r_2dot * self.static_parameters.time_step)  # rad/s      
        for i in range(len(current_state.wheel_w_vel)):
            if current_state.wheel_w_vel[i] <= 0.0:
                current_state.wheel_w_vel[i] = 0.0
                current_state.x_rr.pho_r_2dot[i] = 0.0

        return current_state 



import logging
import sys


class LocalLogger(logging.Logger):
    """Logger util for the package"""
    loggers = set()

    def __init__(self, name, format="%(asctime)s | %(levelname)s | %(funcName)s %(lineno)d | %(message)s", level=logging.INFO):
        # Initial construct.
        self.format = format
        self.level = level
        self.name = name

        # Logger configuration.
        self.console_formatter = logging.Formatter(self.format)
        self.console_logger = logging.StreamHandler(sys.stdout)
        self.console_logger.setFormatter(self.console_formatter)

        # Complete logging config.
        self.logger = logging.getLogger(name)
        if name not in self.loggers:
            self.loggers.add(name)
            self.logger.setLevel(self.level)
            self.logger.addHandler(self.console_logger)



""" 
Main Function for the vehicle dynamics module.
@author: Maikol Funk Drechsler, Yuri Poledna

ROADVIEW Project
Co-Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them. UK and Swiss participants in this project are supported by Innovate UK (contract no. 10045139) and the Swiss State Secretariat for Education, Research and Innovation (Contract no.22.00123) respectevely.
"""
from vehicle_dynamics.structures.OutputStates import OutputStates
from vehicle_dynamics.utils.plot_function import plot_function
from vehicle_dynamics.structures.Manoeuvre import Manoeuvre
from vehicle_dynamics.structures.StateVector import StateVector

from vehicle_dynamics.utils.StaticParameters import StaticParameters

from vehicle_dynamics.VehicleDynamics import VehicleDynamics

import pandas
import numpy as np
import tqdm

from collections import namedtuple

import pickle
import os
import urllib.request
file_path = "example_data/Braking.pickle"
url = "https://github.com/roadview-project/vehicle-dynamics/blob/main/example_data/Braking.pickle"
urllib.request.urlretrieve(url, "Braking.pickle")
os.makedirs(os.path.dirname(file_path), exist_ok=True)

if not os.path.exists(file_path):
    print(f"{file_path} not found. Downloading from {url}...")
    urllib.request.urlretrieve(url, file_path)
    print("Download complete!")
else:
    print(f"{file_path} already exists. Skipping download.")

with open(file_path,"rb") as handle:
    data=pickle.load(handle)

frequency = 1000

steering = data["steering"]
throttle = data["throttle"]
brake = data["brake"]

points = len(brake)
end_time = points/frequency

time = data["time"]

manoeuvre_M8 = Manoeuvre(steering, throttle, brake, time)

output_states = OutputStates()
initial_state = StateVector(x = np.array(data["Rel_pos_x"][0]),
                            y = np.array(data["Rel_pos_y"][0]),
                            vx = np.array(data["Velocity_X"][0]),
                            yaw = np.array(data["Yaw"][0]))

car_parameters_path = "bmw_m8.yaml"

vehicle_dynamics = VehicleDynamics(initial_state = initial_state, 
                                   initial_gear = 1, 
                                   frequency = frequency, 
                                   car_parameters_path = car_parameters_path)
simulation_range = range(0, len(time))

for i in tqdm.tqdm(simulation_range):
    try:
        output_states.set_states(vehicle_dynamics.tick(*manoeuvre_M8[i]))
    except ValueError as e:
        print(e)
        output_states.padding(len(manoeuvre_M8) - len(output_states))
        plot_function(output_states, manoeuvre_M8,data,
                      xlim=(-.25,time[-1] + 0.25),
                      plot_type = "powertrain")
        exit()

plot_function(output_states, manoeuvre_M8,
              xlim=(-.25,time[-1] + 0.25),
              data=data,
              savefig=False,
              fig_save_dir="figures/", 
              static_parameters = StaticParameters(car_parameters_path, frequency),
              initial_iteraction = 0)



"""
Vehicle Dynamic Model - CurrentStates Class

@author:   Maikol Funk Drechsler, Yuri Poledna

Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them.
"""

from vehicle_dynamics.utils.StaticParameters import StaticParameters
from vehicle_dynamics.structures.TireForces import TireForces
from vehicle_dynamics.structures.StateVector import StateVector
from vehicle_dynamics.structures.Displacement import Displacement
from vehicle_dynamics.structures.StrutForce import StrutForce
from vehicle_dynamics.structures.WheelHubForce import WheelHubForce
from vehicle_dynamics.structures.AngularWheelPosition import AngularWheelPosition

from vehicle_dynamics.utils.import_data_CM import import_data_CM
from vehicle_dynamics.utils.LocalLogger import LocalLogger

from scipy.interpolate import interp1d
from numpy.linalg import inv
import numpy as np
import yaml
import math 


class CurrentStates(object):
    """This class initialize the values of a vehicular dynamic model. """

    def __init__(self, static_parameters, freq = 1000, initial_state = np.zeros(15), initial_gear = 1, logger = 1):
        super(CurrentStates, self).__init__()
        self.delta = 0

        self.engine_w = static_parameters.powertrain.engine.idle_rpm
        self.gear = initial_gear                      # gear selector
        self.wheel_w_vel = np.zeros(4)

        # State initiate with the position, orientation and speed provided by the arguments, acc = 0; 
        if isinstance(initial_state, np.ndarray):
            self.x_a = StateVector(*initial_state)  # initial_state
        elif isinstance(initial_state, StateVector):
            self.x_a = initial_state

        self.reference_zCG = self.x_a.z

        Ф = self.x_a.roll
        θ = self.x_a.pitch
        Ψ = self.x_a.yaw
        cos = np.cos
        sin = np.sin
        sin_Ф = self.x_a.sin_roll
        cos_Ф = self.x_a.cos_roll
        sin_θ = self.x_a.sin_pitch
        cos_θ = self.x_a.cos_pitch
        sin_Ψ = self.x_a.sin_yaw
        cos_Ψ = self.x_a.cos_yaw

        # Wheel initiate stoped 
        self.x_rr = AngularWheelPosition(pho_r=np.zeros(4), pho_r_dot = np.zeros(4), pho_r_2dot =np.zeros(4))
        self.x_rf = TireForces(fx =np.zeros(4), fy = np.zeros(4), wheel_forces_transformed_force2vehicle_sys = np.ones((3, 4), dtype=float))
        self.displacement = Displacement(static_suspension=(static_parameters.body.mass * static_parameters.wd * static_parameters.gravity) / static_parameters.suspension.spring_rate, 
                                         suspension = np.zeros(4), 
                                         suspension_dot=np.zeros(4), 
                                         zr_dot=np.zeros(4), 
                                         zr_2dot=np.zeros(4))
        zCG = self.x_a.z
        static_cg_height = static_parameters.body.sz
        long_f = static_parameters.body.lf
        # x-position from Vehicle Cog to the rear axle [m]
        long_r = - static_parameters.body.lr
        # y-position from Vehicle CoG to the left chassis point [m]
        lat_l = static_parameters.body.wl
        # y-position from Vehicle CoG to the right chassis point  [m]
        lat_r = - static_parameters.body.wr

        zCG = self.x_a.z
        static_cg_height = self.reference_zCG

        displacement_FL = zCG - static_cg_height - (sin_θ * long_f) + (sin_Ф * lat_l) - self.displacement.road[0]
        displacement_RL = zCG - static_cg_height - (sin_θ * long_r) + (sin_Ф * lat_l) - self.displacement.road[1]
        displacement_FR = zCG - static_cg_height - (sin_θ * long_f) + (sin_Ф * lat_r) - self.displacement.road[2]
        displacement_RR = zCG - static_cg_height - (sin_θ * long_r) + (sin_Ф * lat_r) - self.displacement.road[3]

        displacement_suspension = np.array([displacement_FL, displacement_RL, displacement_FR, displacement_RR])

        self.displacement.suspension_dot = (displacement_suspension - self.displacement.suspension) / static_parameters.time_step
        self.displacement.suspension = displacement_suspension

        sprung_mass = static_parameters.body.mass * static_parameters.wd - static_parameters.suspension.unsprung_mass

        damping_force = (static_parameters.suspension.damping_rate * self.displacement.suspension_dot)
        spring_force = (static_parameters.suspension.spring_rate * displacement_suspension)

        anti_roll_bar_force = static_parameters.suspension.roll_bar_stiffness * sin_Ф / (2 * np.array([lat_l, lat_l, lat_r, lat_r]))

        suspension_force = spring_force + damping_force - anti_roll_bar_force

        self.f_zr = WheelHubForce(f_zr_dot=np.array([0., 0., 0., 0.]), 
                                  wheel_load_z = static_parameters.body.mass * static_parameters.gravity * static_parameters.wd)

        static_force = static_parameters.wd * static_parameters.body.mass

        lateral_force_transfer = np.array([-1, -1, 1, 1]) * (np.sum(sprung_mass) * self.x_a.acc_y * static_parameters.suspension.roll_centre_height / static_parameters.track_width)

        longitudinal_force_transfer = np.array([-1, 1, -1, 1]) * (np.sum(sprung_mass) * self.x_a.acc_x * static_parameters.suspension.pitch_centre_height / (2 * static_parameters.wheel_base))

        self.f_zr.wheel_load_z = sprung_mass * static_parameters.gravity + suspension_force + longitudinal_force_transfer + lateral_force_transfer

        self.f_za = StrutForce(f_za = static_parameters.body.mass * static_parameters.gravity * static_parameters.wd, 
                               f_za_dot = np.zeros(4), 
                               spring_force = (static_parameters.body.mass * np.array([static_parameters.body.lr * static_parameters.body.wr, static_parameters.body.lf * static_parameters.body.wr, static_parameters.body.lr * static_parameters.body.wl, static_parameters.body.lf * static_parameters.body.wl]) / ((static_parameters.body.lr + static_parameters.body.lf) * (static_parameters.body.wl + static_parameters.body.wr)) * static_parameters.gravity).reshape(-1, 1), 
                               dumper_force = np.zeros((4, 1), dtype=float))

        self.x_rf.wheel_forces_transformed_force2vehicle_sys[2, :] = self.f_zr.wheel_load_z

        self.compiled_wheel_forces = np.array([self.x_rf.fx, self.x_rf.fy, self.f_zr.wheel_load_z])

        # Creating vector for chassis method 
        self.sum_f_wheel = np.zeros(3, dtype=float)  # Sum of wheel forces

        self.slip_x = np.zeros(4)
        self.slip_y = np.zeros(4)
        self.slip_y_rate = np.zeros(4)
        

        self.powertrain_net_torque = np.zeros(4)
        self.suspension_force = suspension_force


    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"delta was set as {value}")
        self._delta = value

    @property
    def engine_w(self):
        return self._engine_w

    @engine_w.setter
    def engine_w(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"engine_w was set as {value}")
        self._engine_w = value

    @property
    def gear(self):
        return self._gear

    @gear.setter
    def gear(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"gear was set as {value}")
        self._gear = value

    @property
    def wheel_w_vel(self):
        return self._wheel_w_vel

    @wheel_w_vel.setter
    def wheel_w_vel(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"wheel_w_vel was set as {value}")
        self._wheel_w_vel = value

    @property
    def x_a(self):
        return self._x_a

    @x_a.setter
    def x_a(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"x_a was set as {value}")
        self._x_a = value

    @property
    def reference_zCG(self):
        return self._reference_zCG

    @reference_zCG.setter
    def reference_zCG(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"reference_zCG was set as {value}")
        self._reference_zCG = value

    @property
    def x_rr(self):
        return self._x_rr

    @x_rr.setter
    def x_rr(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"x_rr was set as {value}")
        self._x_rr = value

    @property
    def x_rf(self):
        return self._x_rf

    @x_rf.setter
    def x_rf(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"x_rf was set as {value}")
        self._x_rf = value

    @property
    def displacement(self):
        return self._displacement

    @displacement.setter
    def displacement(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"displacement was set as {value}")
        self._displacement = value

    @property
    def f_zr(self):
        return self._f_zr

    @f_zr.setter
    def f_zr(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"f_zr was set as {value}")
        self._f_zr = value

    @property
    def f_za(self):
        return self._f_za

    @f_za.setter
    def f_za(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"f_za was set as {value}")
        self._f_za = value

    @property
    def compiled_wheel_forces(self):
        return self._compiled_wheel_forces

    @compiled_wheel_forces.setter
    def compiled_wheel_forces(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"compiled_wheel_forces was set as {value}")
        self._compiled_wheel_forces = value

    @property
    def sum_f_wheel(self):
        return self._sum_f_wheel

    @sum_f_wheel.setter
    def sum_f_wheel(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"sum_f_wheel was set as {value}")
        self._sum_f_wheel = value

    @property
    def slip_x(self):
        return self._slip_x

    @slip_x.setter
    def slip_x(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"slip_x was set as {value}")
        self._slip_x = value

    @property
    def slip_y(self):
        return self._slip_y

    @slip_y.setter
    def slip_y(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"slip_y was set as {value}")
        self._slip_y = value

    @property
    def powertrain_net_torque(self):
        return self._powertrain_net_torque

    @powertrain_net_torque.setter
    def powertrain_net_torque(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"powertrain_net_torque was set as {value}")
        self._powertrain_net_torque = value

    @property
    def suspension_force(self):
        return self._suspension_force

    @suspension_force.setter
    def suspension_force(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"suspension_force was set as {value}")
        self._suspension_force = value


def main():
    from vehicle_dynamics.utils.LocalLogger import LocalLogger
    from vehicle_dynamics.utils.StaticParameters import StaticParameters
    test_function = CurrentStates
    function_name = test_function.__name__
    logger = LocalLogger(function_name).logger
    logger.setLevel('INFO')
    static_parameters = StaticParameters("../../bmw_m8.yaml")
    parameters = CurrentStates(static_parameters, logger=logger)
    pass


if __name__ == '__main__':
    main()



"""
Vehicle Dynamic Model - Body Class

@author:   Maikol Funk Drechsler, Yuri Poledna

Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them.
"""
from vehicle_dynamics.utils.StaticParameters import StaticParameters
from vehicle_dynamics.structures.StateVector import StateVector
from vehicle_dynamics.structures.WheelHubForce import WheelHubForce
from vehicle_dynamics.structures.Displacement import Displacement
from vehicle_dynamics.utils.CurrentStates import CurrentStates
from vehicle_dynamics.utils.import_data_CM import import_data_CM
from vehicle_dynamics.utils.LocalLogger import LocalLogger
from vehicle_dynamics.structures.OutputStates import OutputStates
from vehicle_dynamics.structures.Manoeuvre import Manoeuvre
from vehicle_dynamics.utils.plot_function import plot_function

import numpy as np
import logging
from copy import copy
import yaml


class Body():
    def __init__(self, static_parameters, logger: logging.Logger):
        self.static_parameters = static_parameters
        self.logger = logger

    def access_z_road(self, x, y):
        # TODO implement topography
        z = 0.
        return z

    def road(self, current_state):
        for k in range(4):
            # displacement.zs[k][2] = self.access_z_road(self.displacement.zs[k][0],self.displacement.zs[k][1])
            current_state.displacement.road[k] = 0.

        return current_state 

    def body(self, current_state: CurrentStates):
        """
         body is a function that calculates the current wheel loads (z)

         Required current_state from static_parameters:
             1. suspension_spring_rate
             2. suspension_damping_rate

         Required Arguments:
             1. f_zr
                 1.01 wheel_load_z
            2.displacement
                2.01 za    :  position of the connection point of the suspension with the dampner spring system
                2.02 zs    :  Z velocity of the wheel
                2.03 l_stat:  constant
                2.04 za_dot:  velocity of the chassis point
            3.vehicle_fixed2inertial_system
         Returns:
             1. f_zr.wheel_load_z

         """
        vx = current_state.x_a.vx
        vy = current_state.x_a.vy
        Ф = current_state.x_a.roll
        θ = current_state.x_a.pitch
        Ψ = current_state.x_a.yaw
        sin_Ф = current_state.x_a.sin_roll
        cos_Ф = current_state.x_a.cos_roll
        sin_θ = current_state.x_a.sin_pitch
        cos_θ = current_state.x_a.cos_pitch
        sin_Ψ = current_state.x_a.sin_yaw
        cos_Ψ = current_state.x_a.cos_yaw
        θdot = current_state.x_a.wx
        Фdot = current_state.x_a.wy
        Ψdot = current_state.x_a.wz
        θdotdot = current_state.x_a.wx_dot
        Фdotdot = current_state.x_a.wy_dot
        Ψdotdot = current_state.x_a.wz_dot
        h = self.static_parameters.body.sz
        m = self.static_parameters.body.mass

        sum_f_wheel = np.sum(current_state.x_rf.wheel_forces_transformed_force2vehicle_sys, axis=1)

        current_state = self.road(current_state)  # placeholder to consider road status

        # x-position from Vehicle CoG to the front axle [m]
        long_f = self.static_parameters.body.lf
        # x-position from Vehicle Cog to the rear axle [m]
        long_r = self.static_parameters.body.lr
        # y-position from Vehicle CoG to the left chassis point [m]
        lat_l = self.static_parameters.body.wl
        # y-position from Vehicle CoG to the right chassis point  [m]
        lat_r = self.static_parameters.body.wr

        l = long_r + long_f
        w = lat_l + lat_r


        longitudinal_distance = np.array([long_f, -long_r, long_f, -long_r])
        lateral_distance = np.array([lat_l, lat_l, - lat_r, - lat_r])

        zCG = current_state.x_a.z
        static_cg_height = current_state.reference_zCG

        displacement_suspension = zCG - static_cg_height - (longitudinal_distance * sin_θ) + (lateral_distance * sin_Ф) - current_state.displacement.road

        current_state.displacement.suspension_dot = (displacement_suspension - current_state.displacement.suspension) / self.static_parameters.time_step
        current_state.displacement.suspension = displacement_suspension

        sprung_mass = self.static_parameters.sprung_mass

        damping_force = -(self.static_parameters.suspension.damping_rate * current_state.displacement.suspension_dot)
        spring_force = -(self.static_parameters.suspension.spring_rate * displacement_suspension)

        anti_roll_bar_force = self.static_parameters.suspension.roll_bar_stiffness * sin_Ф / (2 * lateral_distance)

        suspension_force = spring_force + damping_force + anti_roll_bar_force

        current_state.suspension_force = suspension_force

        front_wheels_sum = sprung_mass[0] + sprung_mass[2]
        rear_wheels_sum = sprung_mass[1] + sprung_mass[3]

        static_force = self.static_parameters.wd * self.static_parameters.body.mass


        ξ_lon = np.array([-1/2, 1/2, -1/2, 1/2])
        ξ_lat = np.array([- long_r/w, - long_f/w, long_r/w, long_f/w])

        lateral_force_transfer = ξ_lat * (m * current_state.x_a.acc_y * self.static_parameters.suspension.roll_centre_height / self.static_parameters.track_width)

        longitudinal_force_transfer = ξ_lon * (m* current_state.x_a.acc_x * self.static_parameters.suspension.pitch_centre_height / self.static_parameters.wheel_base)

        current_state.f_zr.wheel_load_z = (sprung_mass+self.static_parameters.suspension.unsprung_mass) * self.static_parameters.gravity + suspension_force + longitudinal_force_transfer + lateral_force_transfer

        current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[2, :] = current_state.f_zr.wheel_load_z 


        # CHASSIS

        fy_fl = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[1, 0]
        fy_rl = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[1, 1]
        fy_fr = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[1, 2]
        fy_rr = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[1, 3]

        fx_fl = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[0, 0]
        fx_rl = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[0, 1]
        fx_fr = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[0, 2]
        fx_rr = current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[0, 3]

        #self.logger.info(sum_f_wheel[0])

        current_state.x_a.acc_x = (sum_f_wheel[0] - 0.5 * self.static_parameters.aerodynamics_front * vx**2) / m 
        current_state.x_a.vx += (current_state.x_a.acc_x + vy * Ψdot) * self.static_parameters.time_step
        
        if current_state.x_a.vx <= 0.0:
            current_state.x_a.acc_x = 0.0
            current_state.x_a.vx = 0.0
        

        current_state.x_a.acc_y = (sum_f_wheel[1]) / m  
        current_state.x_a.acc_z = sum(suspension_force) / m

        current_state.x_a.vy += (current_state.x_a.acc_y - vx * Ψdot) * self.static_parameters.time_step
        current_state.x_a.vz += current_state.x_a.acc_z * self.static_parameters.time_step

        vxx, vyx = vx * cos_Ψ, vy * -sin_Ψ
        vxy, vyy = vx * sin_Ψ, vy * cos_Ψ

        current_state.x_a.x += (vxx + vyx) * self.static_parameters.time_step  # (vxx + vyx)
        current_state.x_a.y += (vxy + vyy) * self.static_parameters.time_step  # (vxy + vyy)
        current_state.x_a.z += current_state.x_a.vz * self.static_parameters.time_step

        # Chassis rotation

        Ix = self.static_parameters.body.i_x_s
        Iy = self.static_parameters.body.i_y_s
        Iz = self.static_parameters.body.i_z
        h = self.static_parameters.body.sz
        l = self.static_parameters.body.wr    # halftrack

        Fx = sum_f_wheel[0]
        Fy = sum_f_wheel[1]

        #Mz = ((fy_fl + fy_fr) * self.static_parameters.body.lf - (fy_rl + fy_rr) * self.static_parameters.body.lr + (fx_rr + fx_fr - fx_rl - fx_fl) * l)
        Mz = ((fy_fl + fy_fr) * self.static_parameters.body.lf - (fy_rl + fy_rr) * self.static_parameters.body.lr + (fx_rr + fx_fr) * (self.static_parameters.body.wr) + (fx_rl + fx_fl) * (-self.static_parameters.body.wl))

        hsr = self.static_parameters.suspension.roll_centre_height - h
        hsp = self.static_parameters.suspension.pitch_centre_height - h
        

        current_state.x_a.wx_dot = (np.sum(lateral_distance * suspension_force) + np.sum(sprung_mass) * hsr * (current_state.x_a.acc_y + self.static_parameters.gravity * sin_Ф)) / Ix
        current_state.x_a.wy_dot = ((-1) * ((np.sum(longitudinal_distance * suspension_force) + np.sum(sprung_mass) * hsp * (current_state.x_a.acc_x - self.static_parameters.gravity * sin_θ)) / Iy))
        
        current_state.x_a.wz_dot = Mz / Iz   #- h * (Fx * sin_Ф + Fy * sin_θ * cos_Ф)) / (Ix * (sin_θ**2) + (cos_θ**2) * (Iy * (sin_Ф**2) + Iz * (cos_Ф**2)))

        current_state.x_a.wx += current_state.x_a.wx_dot * self.static_parameters.time_step
        current_state.x_a.wy += current_state.x_a.wy_dot * self.static_parameters.time_step
        current_state.x_a.wz += (current_state.x_a.wz_dot * self.static_parameters.time_step)

        # Angular position
        current_state.x_a.roll += (current_state.x_a.wx * self.static_parameters.time_step)
        current_state.x_a.pitch += (current_state.x_a.wy * self.static_parameters.time_step)
        current_state.x_a.yaw += (current_state.x_a.wz * self.static_parameters.time_step)

        return current_state


def main(static_parameters, current_state, data, logger, savefig=False):
    import yaml
    from tqdm import tqdm

    body = Body(static_parameters, logger)

    test_function = body.body
    function_name = test_function.__name__

    steering = data["steering"]
    throttle = data["throttle"]
    brake = data["brake"]
    points = len(brake)

    Fx_fl = data["Fx_fl"]
    Fx_fr = data["Fx_fr"]
    Fx_rl = data["Fx_rl"]
    Fx_rr = data["Fx_rr"]

    Fy_fl = data["Fy_fl"]
    Fy_fr = data["Fy_fr"]
    Fy_rl = data["Fy_rl"]
    Fy_rr = data["Fy_rr"]

    time = data["time"]
    manoeuvre = Manoeuvre(steering, throttle, brake, time)
    logger.info("loaded SimulationData")
    simulation_range = range(0, len(time))
    output_states = OutputStates()

    for i in tqdm(simulation_range):
        forces_in_x = [Fx_fl[i], Fx_rl[i], Fx_fr[i], Fx_rr[i]]
        forces_in_y = [Fy_fl[i], Fy_rl[i], Fy_fr[i], Fy_rr[i]]
        current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[0, :] = forces_in_x
        current_state.x_rf.wheel_forces_transformed_force2vehicle_sys[1, :] = forces_in_y

        output_states.set_states(test_function(current_state))
    plot_function(output_states, manoeuvre, data, savefig, 
                  plot_type="body",
                  xlim=(-.25, data["time"][-1] - data["time"][0]+0.25),
                  static_parameters=static_parameters)


if __name__ == '__main__':
    import pickle
    test_function = Body
    function_name = test_function.__name__
    logger = LocalLogger(function_name).logger
    logger.setLevel('INFO')

    PATH_TO_DATA = "../../example_data/Braking.pickle"
    with open(PATH_TO_DATA,"rb")as handle:
        data=pickle.load(handle)

    from vehicle_dynamics.structures.StateVector import StateVector
    from vehicle_dynamics.utils.StaticParameters import StaticParameters
    static_parameters = StaticParameters("../../bmw_m8.yaml", freq = 1000)

    initial_state = StateVector(x = np.array( data["Rel_pos_x"][0]),
                                y = np.array(data["Rel_pos_y"][0]),
                                vx = np.array( data["Velocity_X"][0]),
                                yaw = np.array( data["Yaw"][0]))    
                                #yaw = np.array( (sync_adma[0].ins_yaw+125)* 3.14/180.0))


    current_state = CurrentStates(static_parameters, logger=logger, initial_state=initial_state)
    logger.info("loaded current_state")

    logger.info("loaded SimulationData")
    main(static_parameters, current_state, data, logger, False)



"""
Vehicle Dynamic Model Main Class

@author:   Maikol Funk Drechsler, Yuri Poledna

Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them.
"""

from vehicle_dynamics.structures.TireForces import TireForces
from vehicle_dynamics.structures.StateVector import StateVector
from vehicle_dynamics.structures.Displacement import Displacement

from vehicle_dynamics.structures.WheelHubForce import WheelHubForce
from vehicle_dynamics.structures.AngularWheelPosition import AngularWheelPosition
from vehicle_dynamics.structures.OutputStates import OutputStates
from vehicle_dynamics.structures.CurrentStates import CurrentStates
from vehicle_dynamics.structures.StaticParameters import StaticParameters

from vehicle_dynamics.modules.LocalLogger import LocalLogger

from vehicle_dynamics.modules.powertrain import Powertrain
from vehicle_dynamics.modules.wheels import Wheels
from vehicle_dynamics.modules.body import Body


from scipy.interpolate import interp1d
from numpy.linalg import inv
import numpy as np
import yaml
import math


class VehicleDynamics(object):
    """ This Class does the grunt of calculating VehicleDynamics!!
    Calculate longitudinal and lateral dynamics with desired values 
    of brake, steer and thorttle positons.
    """

    def __init__(self, initial_state = np.zeros(15), initial_gear = 1, frequency=1000, car_parameters_path = ""):
        self.logger = LocalLogger("MainLogger").logger
        self.logger.setLevel("INFO")

        self.static_parameters = StaticParameters(car_parameters_path, frequency)

        self.logger.info("Imported YAML car self")    
        self.current_states = CurrentStates(self.static_parameters, frequency, initial_state, initial_gear, self.logger)
        self.powertrain = Powertrain(self.static_parameters, self.logger)
        self.wheels = Wheels(self.static_parameters, self.logger)
        self.body = Body(self.static_parameters, self.logger)
        self.CUT_VALUE = 0.2

    def tick(self, throttle, brake, steering_angle):
        if self.CUT_VALUE > np.any(self.current_states.slip_x):
            throttle=throttle/2
        self.current_states = self.powertrain.powertrain(self.current_states, throttle, brake)
        self.current_states = self.wheels.wheels(self.current_states, steering_angle)
        self.current_states = self.body.body(self.current_states) 

        return self.current_states



from collections import namedtuple

import numpy as np
ManoeuvreState = namedtuple('ManoeuvreState', 'throttle brake steering')


class Manoeuvre():
    """docstring for Manoeuvre"""

    def __init__(self, steering, throttle, brake, time):
        super(Manoeuvre, self).__init__()
        self.steering = steering
        self.throttle = throttle
        self.brake = brake
        self.time = time

    def __getitem__(self, items):
        """ 
        receives items to be returned
        returns ManoeuvreState(throttle, brake, steering)

        """
        return ManoeuvreState(self.throttle[items], self.brake[items], self.steering[items])

    def __array__(self) -> np.ndarray:
        return np.array([self.steering,
                         self.throttle,
                         self.brake,
                         self.time])

    def __len__(self):
        return len(self.time)
        pass



from vehicle_dynamics.utils.StaticParameters import StaticParameters
from vehicle_dynamics.utils.LocalLogger import LocalLogger
from vehicle_dynamics.utils.CurrentStates import CurrentStates
from vehicle_dynamics.utils.import_data_CM import import_data_CM
from vehicle_dynamics.utils.plot_function import plot_function
from vehicle_dynamics.utils.StaticParameters import StaticParameters

from vehicle_dynamics.structures.OutputStates import OutputStates
from vehicle_dynamics.structures.Manoeuvre import Manoeuvre

import logging
from scipy.interpolate import interp1d
import numpy as np


class Powertrain(object):
    """
        Powertrain is a class calculates the current Torque delivered by the engine to the wheels.
        For that is necessary select the proper gear and find the torque 
        converter multiplicator.
        gear_change() function uses engine angular velocity and car linear speed (vx)
        torque_converter() multiplication is based on input and output rotation ratio.

        Required current_state from static_parameters:
            1.engine_w_table,
            2.torque_max_table,
            3.gear_ratio,
            4.diff,
            5.diff_ni,
            6.transmition_ni,
            7.gear_selection,
            8.engine_inertia,
            9.axel_inertia,
            10.gearbox_inertia,
            11.shaft_inertia,
            12.wheel_inertia,
            13.max_brake_torque,
            14.brake_bias,

        Required Arguments:
            1. throttle
            2. brake
            3. acc_x
            4.gear[t-1]
            5. vx

        Returns: (current_state)
            1. Engine_w
            2. Gear
            3. Torque of the engine on the wheels
            4. prev_gear
            5. current_gear

    """

    def __init__(self, static_parameters: StaticParameters, logger: LocalLogger):
        super(Powertrain, self).__init__()
        self.logger = logger
        self.torque_interpolation = interp1d(static_parameters.powertrain.engine.w_table, static_parameters.powertrain.engine.torque_max)
        self.MIN_GEAR_CHANGE_INTERVAL = static_parameters.powertrain.gearbox.MIN_GEAR_CHANGE_INTERVAL
        self.CONVERTER_SYNC_TIME = static_parameters.powertrain.torque_converter.CONVERTER_SYNC_TIME
        self.current_sync = 0
        self.current_grace_period = 0
        self.static_parameters = static_parameters
        self.μ_funct = interp1d(static_parameters.powertrain.torque_converter.speed_ratio,
                                static_parameters.powertrain.torque_converter.ratio)
        self.k_in_funct = interp1d(np.array(np.linspace(0., 1., len(static_parameters.powertrain.torque_converter.factor))), np.array(static_parameters.powertrain.torque_converter.factor))
        self.previous_throtle = 0

    def gear_change_rpm(self, current_state: CurrentStates, throttle: float):
        if self.current_grace_period > 0:
            self.current_grace_period -= 1
            return False

        # Gearbox up or down shifting
        if current_state.engine_w > self.static_parameters.powertrain.gearbox.gear_max_rpm[round(throttle * 10)][current_state.gear]:
            if current_state.gear + 1 >= self.static_parameters.powertrain.gearbox.gear_ratio.size:
                return False
            current_state.gear = current_state.gear + 1
            self.current_grace_period = self.MIN_GEAR_CHANGE_INTERVAL
            return True

        elif current_state.engine_w < self.static_parameters.powertrain.gearbox.gear_min_rpm[round(throttle * 10)][current_state.gear]:
            if current_state.gear - 1 < 1:
                return False
            self.current_grace_period = self.MIN_GEAR_CHANGE_INTERVAL
            current_state.gear = current_state.gear - 1
            return True        

    def powertrain(self, current_state: CurrentStates, throttle: float, brake: float):
        # Update Converter engine_w
        # Based on page 3 from Generating Proper Integrated Dynamic Models for Vehicle Mobility (Loucas S. Louca)
        # Based on the whell velocity and engine engine_w the torque on the

        if self.gear_change_rpm(current_state, throttle):
            self.current_sync = self.CONVERTER_SYNC_TIME

        if self.current_sync > 0:
            torque_converter_out = 0
            self.current_sync -= 1
            throttle = 0.2

        # Calculate torque provided by the engine based on the engine engine_w
        torque_available = self.torque_interpolation(current_state.engine_w)
        
        if self.previous_throtle - throttle != 0 :
            self.current_grace_period += 2
            self.previous_throtle = throttle
        else:
            self.previous_throtle = throttle
        
        gas_pedal = throttle

        # checking for idle state
        if current_state.engine_w < self.static_parameters.powertrain.engine.idle_rpm:
            #throttle = 1
            current_state.engine_w = self.static_parameters.powertrain.engine.idle_rpm

        # find the torque delivered by the engine
        engine_torque = (throttle * torque_available)

        final_ratio = self.static_parameters.powertrain.gearbox.gear_ratio[current_state.gear] * self.static_parameters.powertrain.differential.ratio

        turbine_w = final_ratio * np.mean(current_state.wheel_w_vel)

        torque_converter_ratio = turbine_w / current_state.engine_w


        if torque_converter_ratio >= self.static_parameters.powertrain.torque_converter.lock_up_ratio:
            # Clutch Mode
            current_state.engine_w = turbine_w
            torque_converter_out = engine_torque
        
        else:
            torque_converter_ratio = 0 if torque_converter_ratio < 0 else torque_converter_ratio
            # Torque Converter
            k_in = self.k_in_funct(torque_converter_ratio)
            μ = self.μ_funct(torque_converter_ratio)

            torque_converter_in = k_in * (current_state.engine_w ** 2)
            torque_converter_out = μ * torque_converter_in

            engine_wdot = (engine_torque - torque_converter_in) / self.static_parameters.powertrain.engine.inertia
            current_state.engine_w = (current_state.engine_w + engine_wdot * self.static_parameters.time_step)

        # Gillespie equation 2-7
        traction_torque = torque_converter_out * final_ratio * self.static_parameters.powertrain.gearbox.efficiency
        
        #if traction_torque > 5000.0:
        #    traction_torque = 5000.0

        wheel_torque = traction_torque * self.static_parameters.powertrain.bias

        #if brake > 0:
        #    wheel_torque = np.zeros(4)

        # --------------------Break Torque -------------------------
        brake_torque = brake * self.static_parameters.brake.max_braking_torque * self.static_parameters.brake.brake_bias

        # -------------------- Total Torque -------------------------
        if np.mean((wheel_torque - brake_torque)) <= 0 and current_state.x_a.vx <= 1e-6:
            current_state.powertrain_net_torque = np.zeros(4)
            self.logger.debug(f"case 1")

        elif current_state.x_a.vx <= 1e-6 and brake > 0:
            # IDLE state
            current_state.powertrain_net_torque = np.zeros(4)
            self.logger.debug(f"case 2")

        else:
            current_state.powertrain_net_torque = wheel_torque - brake_torque 

        # Unecessary but for code safety
        if current_state.engine_w > self.static_parameters.powertrain.engine.maximum_rpm:
            current_state.engine_w = self.static_parameters.powertrain.engine.maximum_rpm

        return current_state


def main(static_parameters, current_state, data, logger, savefig=False):
    import yaml
    from tqdm import tqdm

    steering = data["steering"]
    throttle = data["throttle"]
    brake = data["brake"]

    time = data["time"]

    points = len(brake)

    manoeuvre = Manoeuvre(steering, throttle, brake, time)
    logger.info("loaded SimulationData")


    fl,rl,fr,rr = (data["v_wheel_fl"],
                   data["v_wheel_fr"],
                   data["v_wheel_rl"],
                   data["v_wheel_rr"])
    
    simulation_range = range(0, len(time))

    wheel_w_vel = np.array([[fl[i],rl[i],fr[i],rr[i]] for i in range(len(fl))])
    vx = data ["Velocity_X"]
    vy = data ["Velocity_Y"]
    
    yawrate = data["Yaw_Velocity"]

    powertrain = Powertrain(static_parameters, logger)
    test_function = powertrain.powertrain
    output_states = OutputStates()

    for i in tqdm(simulation_range):
        current_state.x_a.vx = vx[i]
        current_state.wheel_w_vel = wheel_w_vel[i]
        try:
            current_state = test_function(current_state, 
                                          throttle = manoeuvre[i].throttle,
                                          brake = manoeuvre[i].brake)
            output_states.set_states(current_state)
        except ValueError as e:
            print(e)
            output_states.padding(len(manoeuvre) - len(output_states))
            plot_function(output_states, manoeuvre, sync_adma, sync_zgw, savefig, plot_type="powertrain", xlim=(-.25,sync_zgw[-1].timestamp-sync_zgw[0].timestamp + 0.25))
            exit()


    plot_function(output_states, manoeuvre, data, savefig, plot_type="powertrain", xlim=(-.25,time[-1]-time[0] + 0.25))


if __name__ == '__main__':
    import pickle

    PATH_TO_DATA = "../../example_data/Braking.pickle"
    with open(PATH_TO_DATA,"rb")as handle:
        data=pickle.load(handle)

    from vehicle_dynamics.structures.StateVector import StateVector
    from vehicle_dynamics.utils.StaticParameters import StaticParameters
    static_parameters = StaticParameters("../../bmw_m8.yaml", freq = 1000)
    
    function_name = "Powertrain"
    logger = LocalLogger(function_name).logger
    logger.info("loaded current_state")
    from vehicle_dynamics.utils.StaticParameters import StaticParameters
    static_parameters = StaticParameters("../../BMW_M8.yaml")
    from vehicle_dynamics.structures.StateVector import StateVector    

    initial_state = StateVector(x = np.array( data["Rel_pos_x"][0]),
                                y = np.array(data["Rel_pos_y"][0]),
                                vx = np.array( data["Velocity_X"][0]),
                                yaw = np.array( data["Yaw"][0]))    

    current_state = CurrentStates(static_parameters, logger=logger,initial_state=initial_state)
    main(static_parameters, current_state, data, logger, False)



import numpy as np


class Displacement(object):
    def __init__(self, static_suspension, suspension=np.zeros(4), suspension_dot=np.zeros(4), zr_dot=np.zeros(4), zr_2dot = np.zeros(4), road =np.zeros(4)):
        self.static_suspension = static_suspension
        self.suspension = suspension
        self.suspension_dot = suspension_dot
        self.zr_2dot = zr_2dot
        self.zr_dot = zr_dot
        self.road = road

    def __array__(self) -> np.ndarray:
        return np.array([self.static_suspension,
                         self.suspension,
                         self.suspension_dot,
                         self.zr_2dot,
                         self.zr_dot,
                         self.road])



"""
Vehicle Dynamic Model - StaticParameters Structure

@author:   Maikol Funk Drechsler, Yuri Poledna

Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them.
"""
import yaml
import numpy as np
from munch import DefaultMunch
from copy import deepcopy
import munch


class StaticParameters(object):
    
    _initialized=False

    def __init__(self, car_parameters_path, freq=1000, optimization=False):
        super(StaticParameters, self).__init__()
        with open(car_parameters_path, 'r') as file:
            param = yaml.safe_load(file)

        self.convert_YAML2Struct(param)

        self.powertrain.gearbox.gear_max_rpm *= np.pi / 30
        self.powertrain.gearbox.gear_min_rpm *= np.pi / 30
        self.powertrain.engine.w_table *= np.pi / 30
        self.powertrain.engine.minimum_rpm *= np.pi / 30
        self.powertrain.engine.maximum_rpm *= np.pi / 30
        self.powertrain.engine.idle_rpm *= np.pi / 30

        self.wheel_base = self.body.lr + self.body.lf
        self.track_width = self.body.wl + self.body.wr

        wd1 = (self.body.lr / (self.wheel_base)) * (self.body.wr / self.track_width)
        wd2 = (self.body.lf / (self.wheel_base)) * (self.body.wr / self.track_width)
        wd3 = (self.body.lr / (self.wheel_base)) * (self.body.wl / self.track_width)
        wd4 = (self.body.lf / (self.wheel_base)) * (self.body.wl / self.track_width)

        self.wd = np.array([wd1, wd2, wd3, wd4])
        self.tire.dynamic_radius = np.array(self.tire.dynamic_radius)
        self.sprung_mass = self.body.mass * self.wd - self.suspension.unsprung_mass

        self.time_step = 1. / freq

        self.aerodynamics_front = self.aerodynamics.air_drag_coefficient * self.aerodynamics.front_area * self.aerodynamics.air_density 

        self.gravity = 9.81

        if not optimization:
            self._initialized = True


    def __setattr__(self, name, value):
        if self._initialized:
            raise AttributeError("We are not in the saving attributes phase of this code, Do Better")
        else:
            object.__setattr__(self, name, value)

    def convert_YAML2Struct(self, yaml):
        data = DefaultMunch.fromDict(yaml)
        for name in dir(data):
            current_key = getattr(data, name)
            setattr(data, name, self.check_munch_key(current_key))
            setattr(self, name, getattr(data, name))

    def check_munch_key(self, current_key):
        for i in dir(current_key):
            if isinstance(getattr(current_key, i), list):
                setattr(current_key, i, np.array(getattr(current_key, i)))
            elif isinstance(getattr(current_key, i), munch.DefaultMunch):
                setattr(current_key, i, self.check_munch_key(getattr(current_key, i)))
        return current_key


def main():
    car_parameters_path = "../../Audi_r8.yaml"

    car_parameters = StaticParameters(car_parameters_path,optimization=True)

    print(car_parameters.aerodynamics.air_drag_coefficient)
    print(car_parameters.powertrain.torque_converter.efficiency_function)
    print(car_parameters.powertrain.torque_converter.efficiency_function)
    car_parameters.gravity = 1
    print(car_parameters.gravity)
    
    car_parameters._initialized = True

    car_parameters.gravity = 9.8
    print(car_parameters.gravity)

if __name__ == '__main__':
    main()



from vehicle_dynamics.utils.StaticParameters import StaticParameters
from vehicle_dynamics.structures.TireForces import TireForces
from vehicle_dynamics.structures.StateVector import StateVector
from vehicle_dynamics.structures.Displacement import Displacement
from vehicle_dynamics.structures.StrutForce import StrutForce
from vehicle_dynamics.structures.WheelHubForce import WheelHubForce
from vehicle_dynamics.structures.AngularWheelPosition import AngularWheelPosition

from vehicle_dynamics.utils.import_data_CM import import_data_CM
from vehicle_dynamics.utils.LocalLogger import LocalLogger


from scipy.interpolate import interp1d
from numpy.linalg import inv
import numpy as np
import yaml
import math 


class CurrentStates(object):
    """This class initialize the values of a vehicular dynamic model. """

    def __init__(self, static_parameters, freq = 1000, initial_state = np.zeros(15), initial_gear = 1, logger = 1):
        super(CurrentStates, self).__init__()
        self.delta = 0

        self.engine_w = static_parameters.powertrain.engine.idle_rpm
        self.gear = initial_gear                      # gear selector
        self.wheel_w_vel = np.zeros(4)

        # State initiate with the position, orientation and speed provided by the arguments, acc = 0; 
        if isinstance(initial_state, np.ndarray):
            self.x_a = StateVector(*initial_state)  # initial_state
        elif isinstance(initial_state, StateVector):
            self.x_a = initial_state

        self.reference_zCG = self.x_a.z

        Ф = self.x_a.roll
        θ = self.x_a.pitch
        Ψ = self.x_a.yaw
        cos = np.cos
        sin = np.sin
        sin_Ф = self.x_a.sin_roll
        cos_Ф = self.x_a.cos_roll
        sin_θ = self.x_a.sin_pitch
        cos_θ = self.x_a.cos_pitch
        sin_Ψ = self.x_a.sin_yaw
        cos_Ψ = self.x_a.cos_yaw

        # Wheel initiate stoped 
        self.x_rr = AngularWheelPosition(pho_r=np.zeros(4), pho_r_dot = np.zeros(4), pho_r_2dot =np.zeros(4))
        self.x_rf = TireForces(fx =np.zeros(4), fy = np.zeros(4), wheel_forces_transformed_force2vehicle_sys = np.ones((3, 4), dtype=float))
        self.displacement = Displacement(static_suspension=(static_parameters.body.mass * static_parameters.wd * static_parameters.gravity) / static_parameters.suspension.spring_rate, 
                                         suspension = np.zeros(4), 
                                         suspension_dot=np.zeros(4), 
                                         zr_dot=np.zeros(4), 
                                         zr_2dot=np.zeros(4))
        zCG = self.x_a.z
        static_cg_height = static_parameters.body.sz
        long_f = static_parameters.body.lf
        # x-position from Vehicle Cog to the rear axle [m]
        long_r = - static_parameters.body.lr
        # y-position from Vehicle CoG to the left chassis point [m]
        lat_l = static_parameters.body.wl
        # y-position from Vehicle CoG to the right chassis point  [m]
        lat_r = - static_parameters.body.wr

        zCG = self.x_a.z
        static_cg_height = self.reference_zCG

        displacement_FL = zCG - static_cg_height - (sin_θ * long_f) + (sin_Ф * lat_l) - self.displacement.road[0]
        displacement_RL = zCG - static_cg_height - (sin_θ * long_r) + (sin_Ф * lat_l) - self.displacement.road[1]
        displacement_FR = zCG - static_cg_height - (sin_θ * long_f) + (sin_Ф * lat_r) - self.displacement.road[2]
        displacement_RR = zCG - static_cg_height - (sin_θ * long_r) + (sin_Ф * lat_r) - self.displacement.road[3]

        displacement_suspension = np.array([displacement_FL, displacement_RL, displacement_FR, displacement_RR])

        self.displacement.suspension_dot = (displacement_suspension - self.displacement.suspension) / static_parameters.time_step
        self.displacement.suspension = displacement_suspension

        sprung_mass = static_parameters.body.mass * static_parameters.wd - static_parameters.suspension.unsprung_mass

        damping_force = (static_parameters.suspension.damping_rate * self.displacement.suspension_dot)
        spring_force = (static_parameters.suspension.spring_rate * displacement_suspension)

        anti_roll_bar_force = static_parameters.suspension.roll_bar_stiffness * sin_Ф / (2 * np.array([lat_l, lat_l, lat_r, lat_r]))

        suspension_force = spring_force + damping_force - anti_roll_bar_force

        self.f_zr = WheelHubForce(f_zr_dot=np.array([0., 0., 0., 0.]), 
                                  wheel_load_z = static_parameters.body.mass * static_parameters.gravity * static_parameters.wd)

        static_force = static_parameters.wd * static_parameters.body.mass

        lateral_force_transfer = np.array([-1, -1, 1, 1]) * (np.sum(sprung_mass) * self.x_a.acc_y * static_parameters.suspension.roll_centre_height / static_parameters.track_width)

        longitudinal_force_transfer = np.array([-1, 1, -1, 1]) * (np.sum(sprung_mass) * self.x_a.acc_x * static_parameters.suspension.pitch_centre_height / (2 * static_parameters.wheel_base))

        self.f_zr.wheel_load_z = sprung_mass * static_parameters.gravity + suspension_force + longitudinal_force_transfer + lateral_force_transfer

        self.f_za = StrutForce(f_za = static_parameters.body.mass * static_parameters.gravity * static_parameters.wd, 
                               f_za_dot = np.zeros(4), 
                               spring_force = (static_parameters.body.mass * np.array([static_parameters.body.lr * static_parameters.body.wr, static_parameters.body.lf * static_parameters.body.wr, static_parameters.body.lr * static_parameters.body.wl, static_parameters.body.lf * static_parameters.body.wl]) / ((static_parameters.body.lr + static_parameters.body.lf) * (static_parameters.body.wl + static_parameters.body.wr)) * static_parameters.gravity).reshape(-1, 1), 
                               dumper_force = np.zeros((4, 1), dtype=float))

        self.x_rf.wheel_forces_transformed_force2vehicle_sys[2, :] = self.f_zr.wheel_load_z

        self.compiled_wheel_forces = np.array([self.x_rf.fx, self.x_rf.fy, self.f_zr.wheel_load_z])

        # Creating vector for chassis method 
        self.sum_f_wheel = np.zeros(3, dtype=float)  # Sum of wheel forces

        self.slip_x = np.zeros(4)
        self.slip_y = np.zeros(4)
        self.slip_y_rate = np.zeros(4)
        

        self.powertrain_net_torque = np.zeros(4)
        self.suspension_force = suspension_force


    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"delta was set as {value}")
        self._delta = value

    @property
    def engine_w(self):
        return self._engine_w

    @engine_w.setter
    def engine_w(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"engine_w was set as {value}")
        self._engine_w = value

    @property
    def gear(self):
        return self._gear

    @gear.setter
    def gear(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"gear was set as {value}")
        self._gear = value

    @property
    def wheel_w_vel(self):
        return self._wheel_w_vel

    @wheel_w_vel.setter
    def wheel_w_vel(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"wheel_w_vel was set as {value}")
        self._wheel_w_vel = value

    @property
    def x_a(self):
        return self._x_a

    @x_a.setter
    def x_a(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"x_a was set as {value}")
        self._x_a = value

    @property
    def reference_zCG(self):
        return self._reference_zCG

    @reference_zCG.setter
    def reference_zCG(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"reference_zCG was set as {value}")
        self._reference_zCG = value

    @property
    def x_rr(self):
        return self._x_rr

    @x_rr.setter
    def x_rr(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"x_rr was set as {value}")
        self._x_rr = value

    @property
    def x_rf(self):
        return self._x_rf

    @x_rf.setter
    def x_rf(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"x_rf was set as {value}")
        self._x_rf = value

    @property
    def displacement(self):
        return self._displacement

    @displacement.setter
    def displacement(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"displacement was set as {value}")
        self._displacement = value

    @property
    def f_zr(self):
        return self._f_zr

    @f_zr.setter
    def f_zr(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"f_zr was set as {value}")
        self._f_zr = value

    @property
    def f_za(self):
        return self._f_za

    @f_za.setter
    def f_za(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"f_za was set as {value}")
        self._f_za = value

    @property
    def compiled_wheel_forces(self):
        return self._compiled_wheel_forces

    @compiled_wheel_forces.setter
    def compiled_wheel_forces(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"compiled_wheel_forces was set as {value}")
        self._compiled_wheel_forces = value

    @property
    def sum_f_wheel(self):
        return self._sum_f_wheel

    @sum_f_wheel.setter
    def sum_f_wheel(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"sum_f_wheel was set as {value}")
        self._sum_f_wheel = value

    @property
    def slip_x(self):
        return self._slip_x

    @slip_x.setter
    def slip_x(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"slip_x was set as {value}")
        self._slip_x = value

    @property
    def slip_y(self):
        return self._slip_y

    @slip_y.setter
    def slip_y(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"slip_y was set as {value}")
        self._slip_y = value

    @property
    def powertrain_net_torque(self):
        return self._powertrain_net_torque

    @powertrain_net_torque.setter
    def powertrain_net_torque(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"powertrain_net_torque was set as {value}")
        self._powertrain_net_torque = value

    @property
    def suspension_force(self):
        return self._suspension_force

    @suspension_force.setter
    def suspension_force(self, value):
        if np.isnan(value).any() or np.isinf(value).any():
            raise ValueError(f"suspension_force was set as {value}")
        self._suspension_force = value


def main():
    from vehicle_dynamics.utils.LocalLogger import LocalLogger
    from vehicle_dynamics.utils.StaticParameters import StaticParameters
    test_function = CurrentStates
    function_name = test_function.__name__
    logger = LocalLogger(function_name).logger
    logger.setLevel('INFO')
    static_parameters = StaticParameters("../../Audi_r8.yaml")
    parameters = CurrentStates(static_parameters, logger=logger)
    pass


if __name__ == '__main__':
    main()



import numpy as np


class WheelHubForce(object):
    def __init__(self, f_zr_dot=np.zeros(4), wheel_load_z =np.array([2000., 3000., 2000., 3000.])):
        self.f_zr_dot = f_zr_dot
        self.wheel_load_z = wheel_load_z

    def __array__(self) -> np.ndarray:
        return np.array([self.f_zr_dot,
                         self.wheel_load_z])



"""
Vehicle Dynamic Model - Logger Class

@author:   Maikol Funk Drechsler, Yuri Poledna

Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them.
"""
import logging
import sys


class LocalLogger(logging.Logger):
    """Logger util for the package"""
    loggers = set()

    def __init__(self, name, format="%(asctime)s | %(levelname)s | %(funcName)s %(lineno)d | %(message)s", level=logging.INFO):
        # Initial construct.
        self.format = format
        self.level = level
        self.name = name

        # Logger configuration.
        self.console_formatter = logging.Formatter(self.format)
        self.console_logger = logging.StreamHandler(sys.stdout)
        self.console_logger.setFormatter(self.console_formatter)

        # Complete logging config.
        self.logger = logging.getLogger(name)
        if name not in self.loggers:
            self.loggers.add(name)
            self.logger.setLevel(self.level)
            self.logger.addHandler(self.console_logger)



import numpy as np


class AngularWheelPosition(object):
    def __init__(self, pho_r=np.zeros(4), pho_r_dot = np.zeros(4), pho_r_2dot = np.zeros(4)):

        # pag 272 eq.52
        # Angular position/speed of each wheel
        self.pho_r = pho_r
        self.pho_r_dot = pho_r_dot
        self.pho_r_2dot = pho_r_2dot

    def __array__(self) -> np.ndarray:
        return np.array([self.pho_r, self.pho_r_dot, self.pho_r_2dot])



"""
Vehicle Dynamic Model - Plot Function

@author:   Maikol Funk Drechsler, Yuri Poledna

Funded by the European Union (grant no. 101069576). Views and opinions expressed are however those of the author(s) only and do not necessarily reflect those of the European Union or the European Climate, Infrastructure and Environment Executive Agency (CINEA). Neither the European Union nor the granting authority can be held responsible for them.
"""
def define_colors():
    import matplotlib.pyplot as plt
    import matplotlib.colors as mcolors
    ROADVIEW_colors = {"ROADVIEW1": "#64798C",
                       "ROADVIEW2": "#B0C6D9",
                       "ROADVIEW3": "#F2C879",
                       "ROADVIEW4": "#F28066"}

    mcolors.get_named_colors_mapping().update(ROADVIEW_colors)
    return mcolors.LinearSegmentedColormap.from_list("", ["ROADVIEW1", "ROADVIEW2", "ROADVIEW3"])


def plot_function(output_states=False, manoeuvre=True, data=True, savefig=False, plot_type = "all", xlim=(), fig_save_dir="../../figures",static_parameters=0, initial_iteraction=0):
    import matplotlib.pyplot as plt
    import numpy as np
    size = 17
    figsize = (size, size*.588)
    simulation_range = range(0, len(manoeuvre))
    throttle = manoeuvre[:].throttle
    plt.rcParams.update({'font.size': 16})

    define_colors()

    vehicle_time = data["car_time"]
    if output_states:
        states = output_states[:].x_a
        if "body" in plot_type or plot_type == "all":
            Fx_fl = data["Fx_fl"]
            Fx_fr = data["Fx_fr"]
            Fx_rl = data["Fx_rl"]
            Fx_rr = data["Fx_rr"]

            Fy_fl = data["Fy_fl"]
            Fy_fr = data["Fy_fr"]
            Fy_rl = data["Fy_rl"]
            Fy_rr = data["Fy_rr"]
            plt.figure(figsize=figsize)
            name = " Chassis Angles"
            plt.title(name)
            var_name = "Vhcl_Roll"
            yaw_vehicle=np.array(data["Yaw"])
            plt.plot(vehicle_time,data["Roll"], "ROADVIEW1", label="M8 Roll")
            plt.plot(vehicle_time,data["Pitch"], "ROADVIEW2", label="M8 Pitch")
            plt.plot(vehicle_time, yaw_vehicle - yaw_vehicle[0], "ROADVIEW3", label="M8 Yaw")
            plt.plot(manoeuvre.time, [i.roll for i in output_states[:].x_a], "ROADVIEW1", linestyle="--", label="Calculated Roll")
            plt.plot(manoeuvre.time, [i.pitch for i in output_states[:].x_a], "ROADVIEW2", linestyle="--", label="Calculated Pitch")
            yaw = np.array([i.yaw for i in output_states[:].x_a])
            plt.plot(manoeuvre.time, yaw - yaw[0], "ROADVIEW3", linestyle="--", label="Calculated Yaw")    
            plt.grid()
            plt.legend()
            plt.xlim(xlim)
            plt.xlabel("Time (s)")
            plt.ylabel("Angle (rad)")
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/" + name + ".png")  
            plt.figure(figsize=figsize)
            name = " Chassis Angles Velocity"
            plt.title(name)
            plt.plot(manoeuvre.time, [i.wx for i in output_states[:].x_a], "ROADVIEW1", linestyle="--", label="Calculated Roll Velocity")
            plt.plot(manoeuvre.time, [i.wy for i in output_states[:].x_a], "ROADVIEW2", linestyle="--", label="Calculated Pitch Velocity")
            plt.plot(manoeuvre.time, [i.wz for i in output_states[:].x_a], "ROADVIEW3", linestyle="--", label="Calculated Yaw Velocity")

            plt.plot(vehicle_time,data["Roll_Velocity"], "ROADVIEW1", label="M8 Roll Velocity")
            plt.plot(vehicle_time,data["Pitch_Velocity"], "ROADVIEW2", label="M8 Pitch Velocity")
            plt.plot(vehicle_time,data["Yaw_Velocity"], "ROADVIEW3", label="M8 Yaw Velocity")
            plt.xlabel("Time (s)")
            plt.ylabel("Angle Velocity (rad/s)")
            plt.legend()
            plt.xlim(xlim)
            plt.grid()
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/" + name + ".png")   

            plt.figure(figsize=figsize)
            name = " Chassis Acceleration"
            plt.title(name)
            plt.plot(vehicle_time, data["Acceleration_X"], "ROADVIEW1", label="M8 Acceleration X")
            plt.plot(vehicle_time, data["Acceleration_Y"], "ROADVIEW2", label="M8 Acceleration Y")
            plt.plot(vehicle_time, data["Acceleration_Z"], "ROADVIEW3", label="M8 Acceleration Z")
            plt.plot(manoeuvre.time, [i.acc_x for i in output_states[:].x_a], "ROADVIEW1", linestyle="--", label="Calculated Acceleration X")
            plt.plot(manoeuvre.time, [i.acc_y for i in output_states[:].x_a], "ROADVIEW2", linestyle="--", label="Calculated Acceleration Y")
            plt.plot(manoeuvre.time, [i.acc_z for i in output_states[:].x_a], "ROADVIEW3", linestyle="--", label="Calculated Acceleration Z")
            plt.grid()
            plt.xlabel("Time (s)")
            plt.ylabel("Acceleration (m/s²)")
            plt.legend()
            plt.xlim(xlim)
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/" + name + ".png")     
            plt.figure(figsize=figsize)
            name = " Chassis Velocity"
            plt.title(name)
            plt.plot(manoeuvre.time, data["Velocity_X"], "ROADVIEW1", label="M8 Velocity X")
            plt.plot(manoeuvre.time, data["Velocity_Y"], "ROADVIEW2", label="M8 Velocity Y")
            plt.plot(manoeuvre.time, data["Velocity_Z"], "ROADVIEW3", label="M8 Velocity Z")

            plt.plot(manoeuvre.time, [i.vx for i in output_states[:].x_a], "ROADVIEW1", linestyle="--", label="Calculated Velocity X")
            plt.plot(manoeuvre.time, [i.vy for i in output_states[:].x_a], "ROADVIEW2", linestyle="--", label="Calculated Velocity Y")
            plt.plot(manoeuvre.time, [i.vz for i in output_states[:].x_a], "ROADVIEW3", linestyle="--", label="Calculated Velocity Z")

            plt.grid()
            plt.xlabel("Time (s)")
            plt.ylabel("Velocity (m/s)")
            plt.legend()
            plt.xlim(xlim)
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/" + name + ".png")   

            plt.figure(figsize=figsize)
            name = " Chassis Position"
            plt.title(name)
            plt.plot(data["Rel_pos_x"],data["Rel_pos_y"], "ROADVIEW1", label="M8 Position")

            plt.plot([i.x for i in output_states[:].x_a], [i.y for i in output_states[:].x_a], "ROADVIEW4", linestyle="--", label="Calculated Position")

            plt.grid()
            plt.xlabel("Position X (m)")
            plt.ylabel("Position Y (m)")
            plt.legend()
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/" + name + ".png")   

        if "powertrain" in plot_type or plot_type == "all":
            gear_changes_idx =np.where(np.roll(output_states[:].gear, 1) != output_states[:].gear)[0]
            gear_changes = manoeuvre.time[gear_changes_idx]

            f, (a0, a1) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [2, 2]}, figsize=figsize, sharex=True)
            powertrain_net_torque = [np.sum(i) for i in output_states[:].powertrain_net_torque]
            a0.plot(manoeuvre.time, powertrain_net_torque, "ROADVIEW1", linestyle="-", label="Calculated")

            powertrain_net_torque_M8 = data["powertrain_net_torque"]
            
            a0.plot(vehicle_time, powertrain_net_torque_M8, "ROADVIEW2",  linestyle="--", label = "BMW M8")
            if len(gear_changes)>0:
                a0.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment")
                a1.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment") 
                for j,i in enumerate(gear_changes[1:]):
                    a0.axvline(i, color="ROADVIEW4", linestyle=":")
                    a1.axvline(i, color="ROADVIEW4", linestyle=":") 
            a0.legend()
            a0.grid()
            a0.set_ylabel("Powertrain Net Torque (Nm)")
            a0.set_xlim(xlim)
            a1.plot(manoeuvre.time, manoeuvre.throttle, "ROADVIEW3", linestyle = "-", label = "Gas pedal(BMW M8)")
            a1.plot(manoeuvre.time, manoeuvre.brake, "ROADVIEW4", linestyle = "-", label = "Brake pedal(BMW M8)")
            a1.step(manoeuvre.time, output_states[:].gear, "ROADVIEW1", linestyle = "-", label = "Calculated Gear Number")
            a1.legend(loc=1)
            a1.set_xlabel("Time (s)")
            a1.grid()
            a1.set_xlim(xlim)
            a1.set_ylim([-0.1, 2.3])
            plt.subplots_adjust(wspace = 0, hspace = 0)
            if savefig:
                plt.savefig(fig_save_dir + "/Powertrain Nettorque.png")

            f, (a0, a1) = plt.subplots(2, 1, gridspec_kw = {'height_ratios': [2, 2]}, figsize = figsize, sharex=True)

            a0.plot(manoeuvre.time, (np.array(output_states[:].engine_w) * 30 / np.pi), "ROADVIEW1", linestyle = "--", label = "Calculated")
            a0.set_ylabel("Engine Rotation (RPM)")

            a0.plot(vehicle_time, data["rpm"], "ROADVIEW2", label = "BMW M8")
            a0.legend()
            a0.grid()
            a0.set_xlim(xlim)
            
            a1.plot(vehicle_time, data["gear"], "ROADVIEW2", linestyle = "-", label = "Gear Number (BMW M8)")
            a1.step(manoeuvre.time, output_states[:].gear, "ROADVIEW1", linestyle = "-", label = "Gear Number (Calculated)")
            a1.plot(manoeuvre.time, manoeuvre.throttle, "ROADVIEW3", linestyle = "-", label = "Gas pedal (BMW M8)")
            a1.plot(manoeuvre.time, manoeuvre.brake, "ROADVIEW4", linestyle = "-", label = "Brake pedal (BMW M8)")
            a1.set_xlim(xlim)
            a1.set_ylim([-0.1, 2.3])
            a1.grid()
            if len(gear_changes)>0:
                a0.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment")
                a1.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment") 
                for j,i in enumerate(gear_changes[1:]):
                    a0.axvline(i, color="ROADVIEW4", linestyle=":")
                    a1.axvline(i, color="ROADVIEW4", linestyle=":") 
            a1.legend(loc =1)
            a1.set_xlabel("Time (s)")
            plt.subplots_adjust(wspace = 0, hspace = 0) 
            if savefig:
                plt.savefig(fig_save_dir + "/Powertrain RPM.png")

        if "wheels" in plot_type or plot_type == "all":


            gear_changes_idx =np.where(np.roll(output_states[:].gear, 1) != output_states[:].gear)[0]
            gear_changes = manoeuvre.time[gear_changes_idx]

            f, (a0, a1) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [3, 1]}, figsize=figsize, sharex=True)
            powertrain_net_torque = [np.sum(i) for i in output_states[:].powertrain_net_torque]
            a0.plot(manoeuvre.time, powertrain_net_torque, "ROADVIEW3", linestyle="--", label="Calculated")

            
            powertrain_net_torque_M8 =  np.array(data["powertrain_net_torque"])
            a0.plot(vehicle_time, powertrain_net_torque_M8, "ROADVIEW1", label = "BMW M8")
            if len(gear_changes)>0:
                a0.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment")
                a1.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment") 
                for j,i in enumerate(gear_changes[1:]):
                    a0.axvline(i, color="ROADVIEW4", linestyle=":")
                    a1.axvline(i, color="ROADVIEW4", linestyle=":") 
            a0.legend()
            a0.grid()
            a0.set_ylabel("Powertrain Net Torque (Nm)")
            a0.set_xlim(xlim)
            a1.plot(manoeuvre.time, manoeuvre.throttle, "ROADVIEW3", linestyle = "--", label = "Gas pedal(BMW M8)")
            a1.plot(manoeuvre.time, manoeuvre.brake, "ROADVIEW4", linestyle = "--", label = "Brake pedal(BMW M8)")
            a1.step(manoeuvre.time, output_states[:].gear, "ROADVIEW1", linestyle = "-", label = "Calculated Gear Number")
            a1.legend(loc=1)
            a1.set_xlabel("Time (s)")
            a1.grid()
            a1.set_xlim(xlim)
            plt.subplots_adjust(wspace = 0, hspace = 0)
            if savefig:
                plt.savefig(fig_save_dir + "/Powertrain Nettorque.png")

            f, (a0, a1) = plt.subplots(2, 1, gridspec_kw = {'height_ratios': [3, 1]}, figsize = figsize, sharex=True)

            a0.plot(manoeuvre.time, (np.array(output_states[:].engine_w) * 30 / np.pi), "ROADVIEW1", linestyle = "--", label = "Calculated")
            a0.set_ylabel("Engine Rotation (RPM)")

            a0.plot(vehicle_time, data["rpm"], "ROADVIEW2", label = "BMW M8")
            a0.legend()
            a0.grid()
            a0.set_xlim(xlim)
            
            a1.plot(vehicle_time, data["gear"], "ROADVIEW2", linestyle = "-", label = "Gear Number (BMW M8)")
            a1.step(manoeuvre.time, output_states[:].gear, "ROADVIEW1", linestyle = "-", label = "Gear Number (Calculated)")
            a1.plot(manoeuvre.time, manoeuvre.throttle, "ROADVIEW3", linestyle = "-", label = "Gas pedal (BMW M8)")
            a1.plot(manoeuvre.time, manoeuvre.brake, "ROADVIEW4", linestyle = "-", label = "Brake pedal (BMW M8)")
            a1.set_xlim(xlim)
            a1.grid()
            if len(gear_changes)>0:
                a0.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment")
                a1.axvline(gear_changes[0], color="ROADVIEW4", linestyle=":", label="Gear Change Moment") 
                for j,i in enumerate(gear_changes[1:]):
                    a0.axvline(i, color="ROADVIEW4", linestyle=":")
                    a1.axvline(i, color="ROADVIEW4", linestyle=":") 
            a1.legend(loc =1)
            a1.set_xlabel("Time (s)")
            plt.subplots_adjust(wspace = 0, hspace = 0) 
            if savefig:
                plt.savefig(fig_save_dir + "/Powertrain RPM.png")
                

            vx = data["Velocity_X"]
            MINIMUM_SPEED_VALUE = 10


            slip_fr = np.zeros(len(vx))
            slip_rr = np.zeros(len(vx)) 
            slip_fl = np.zeros(len(vx))
            slip_rl = np.zeros(len(vx))

            v_wheel_fl = data["v_wheel_fl"]
            v_wheel_fr = data["v_wheel_fr"]
            v_wheel_rl = data["v_wheel_rl"]
            v_wheel_rr = data["v_wheel_rr"]


            for i in range(len(vx)):
              
                if (abs(static_parameters.tire.dynamic_radius[0] * v_wheel_fl[i]) <= MINIMUM_SPEED_VALUE) and (abs(vx[i]) <= MINIMUM_SPEED_VALUE):
                    slip_fl[i] = ((((static_parameters.tire.dynamic_radius[0] * v_wheel_fl[i]) - vx[i]) / (MINIMUM_SPEED_VALUE)))        
                    slip_fl[i] = ((((static_parameters.tire.dynamic_radius[0] * v_wheel_fl[i]) - vx[i]) / np.maximum(np.absolute(static_parameters.tire.dynamic_radius[0] * v_wheel_fl[i]), np.absolute(vx[i]))))

                if (abs(static_parameters.tire.dynamic_radius[0] * v_wheel_fr[i]) <= MINIMUM_SPEED_VALUE) and (abs(vx[i]) <= MINIMUM_SPEED_VALUE):
                    slip_fr[i] = ((((static_parameters.tire.dynamic_radius[0] * v_wheel_fr[i]) - vx[i]) / (MINIMUM_SPEED_VALUE)))        
                    slip_fr[i] = ((((static_parameters.tire.dynamic_radius[0] * v_wheel_fr[i]) - vx[i]) / np.maximum(np.absolute(static_parameters.tire.dynamic_radius[0] * v_wheel_fr[i]), np.absolute(vx[i]))))

                if (abs(static_parameters.tire.dynamic_radius[1] * v_wheel_rl[i]) <= MINIMUM_SPEED_VALUE) and (abs(vx[i]) <= MINIMUM_SPEED_VALUE):
                    slip_rl[i] = ((((static_parameters.tire.dynamic_radius[1] * v_wheel_rl[i]) - vx[i]) / (MINIMUM_SPEED_VALUE)))       
                    slip_rl[i] = ((((static_parameters.tire.dynamic_radius[1] * v_wheel_rl[i]) - vx[i]) / np.maximum(np.absolute(static_parameters.tire.dynamic_radius[1] * v_wheel_rl[i]), np.absolute(vx[i]))))

                if (abs(static_parameters.tire.dynamic_radius[1] * v_wheel_rr[i]) <= MINIMUM_SPEED_VALUE) and (abs(vx[i]) <= MINIMUM_SPEED_VALUE):
                    slip_rr[i] = ((((static_parameters.tire.dynamic_radius[1] * v_wheel_rr[i]) - vx[i]) / (MINIMUM_SPEED_VALUE)))       
                    slip_rr[i] = ((((static_parameters.tire.dynamic_radius[1] * v_wheel_rr[i]) - vx[i]) / np.maximum(np.absolute(static_parameters.tire.dynamic_radius[1] * v_wheel_rr[i]), np.absolute(vx[i]))))

            f, ((a0, a1),(a2,a3)) = plt.subplots(2, 2, figsize=figsize, sharex=True,sharey=True)
            a0.plot(manoeuvre.time, slip_fl, "ROADVIEW1", label="Longitudinal Slip FL")
            a0.plot(manoeuvre.time, [i[0] for i in output_states[:].slip_x], "ROADVIEW1", linestyle ="--", label ="Calculated Longitudinal Slip FL")
            a0.set_xlim(xlim)
            a0.legend()
            a0.set_ylim(-0.2, 0.2)
            a0.set_xlabel("Time (s)")
            a0.set_ylabel("Wheel Slip")
            a2.set_xlabel("Time (s)")
            a2.set_ylabel("Wheel Slip")
            a2.plot(manoeuvre.time, slip_rl, "ROADVIEW2", label="Longitudinal Slip RL")
            a2.plot(manoeuvre.time, [i[1] for i in output_states[:].slip_x], "ROADVIEW2", linestyle ="--", label ="Calculated Longitudinal Slip RL")
            a2.set_xlim(xlim)
            a2.legend()
            a1.set_ylim(-0.2, 0.2)
            a1.set_xlabel("Time (s)")
            a1.set_ylabel("Wheel Slip")
            a1.plot(manoeuvre.time, slip_fr, "ROADVIEW3", label="Longitudinal Slip FR")
            a1.plot(manoeuvre.time, [i[2] for i in output_states[:].slip_x], "ROADVIEW3", linestyle ="--", label ="Calculated Longitudinal Slip FR")
            a1.set_xlim(xlim)
            a1.legend()
            a1.set_ylim(-0.2, 0.2)
            a3.set_ylim(-0.2, 0.2)
            a3.set_xlabel("Time (s)")
            a3.set_ylabel("Wheel Slip")
            a3.plot(manoeuvre.time, slip_rr, "ROADVIEW4", label="Longitudinal Slip RR")
            a3.plot(manoeuvre.time, [i[3] for i in output_states[:].slip_x], "ROADVIEW4", linestyle ="--", label ="Calculated Longitudinal Slip RR")
            a3.set_xlim(xlim)
            a3.legend()
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/Unoptimized wheel_slip_long_subplot-wheelLoad.png")

            f, ((a0, a1),(a2,a3)) = plt.subplots(2, 2, figsize=figsize, sharex=True,sharey=True)
            a0.plot(vehicle_time, data["slip_FL"], "ROADVIEW1", label="Lateral Slip FL")
            a0.plot(manoeuvre.time, [i[0] for i in output_states[:].slip_y], "ROADVIEW1", linestyle ="--", label ="Calculated Lateral Slip FL")
            a0.set_xlim(xlim)
            a0.legend()
            a0.set_ylim(-0.2, 0.2)
            a0.set_xlabel("Time (s)")
            a0.set_ylabel("Wheel Slip")
            a2.set_xlabel("Time (s)")
            a2.set_ylabel("Wheel Slip")
            a2.plot(vehicle_time, data["slip_RL"], "ROADVIEW2", label="Lateral Slip RL")
            a2.plot(manoeuvre.time, [i[1] for i in output_states[:].slip_y], "ROADVIEW2", linestyle ="--", label ="Calculated Lateral Slip RL")
            a2.set_xlim(xlim)
            a2.legend()
            a1.set_ylim(-0.2, 0.2)
            a1.set_xlabel("Time (s)")
            a1.set_ylabel("Wheel Slip")
            a1.plot(vehicle_time, data["slip_FR"], "ROADVIEW3", label="Lateral Slip FR")
            a1.plot(manoeuvre.time, [i[2] for i in output_states[:].slip_y], "ROADVIEW3", linestyle ="--", label ="Calculated Lateral Slip FR")
            a1.set_xlim(xlim)
            a1.legend()
            a1.set_ylim(-0.2, 0.2)
            a3.set_ylim(-0.2, 0.2)
            a3.set_xlabel("Time (s)")
            a3.set_ylabel("Wheel Slip")
            a3.plot(vehicle_time, data["slip_RR"], "ROADVIEW4", label="Lateral Slip RR")
            a3.plot(manoeuvre.time, [i[3] for i in output_states[:].slip_y], "ROADVIEW4", linestyle ="--", label ="Calculated Lateral Slip RR")
            a3.set_xlim(xlim)
            a3.legend()
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/Unoptimized wheel_slip_side_subplot-wheelLoad.png")
            
            plt.figure(figsize=figsize)
            plt.plot(vehicle_time, data["slip_FL"], "ROADVIEW1", label="Lateral Slip FL")
            plt.plot(vehicle_time, data["slip_RL"], "ROADVIEW2", label="Lateral Slip RL")
            plt.plot(vehicle_time, data["slip_FR"], "ROADVIEW3", label="Lateral Slip FR")
            plt.plot(vehicle_time, data["slip_RR"], "ROADVIEW4", label="Lateral Slip RR")
            plt.legend()
            plt.tight_layout()  
            if savefig:
                plt.savefig(fig_save_dir + "/Unoptimized wheel_slip_side_subplot-wheelLoad.png")

            f, ((a0, a1),(a2,a3)) = plt.subplots(2, 2, figsize=figsize, sharex=True,sharey=True)
            a0.plot(manoeuvre.time,  data["v_wheel_fl"], "ROADVIEW1", label="Wheel Velocity FL")
            a0.plot(manoeuvre.time, [i[0] for i in output_states[:].wheel_w_vel], "ROADVIEW1", linestyle ="--", label ="Calculated Wheel Velocity FL")
            a0.set_xlim(xlim)
            a0.legend()
            a0.set_xlabel("Time (s)")
            a0.set_ylabel("Wheel Velocity")
            a2.set_xlabel("Time (s)")
            a2.set_ylabel("Wheel Velocity")
            a2.plot(manoeuvre.time, data["v_wheel_rl"], "ROADVIEW2", label="Wheel Velocity RL")
            a2.plot(manoeuvre.time, [i[1] for i in output_states[:].wheel_w_vel], "ROADVIEW2", linestyle ="--", label ="Calculated Wheel Velocity RL")
            a2.set_xlim(xlim)
            a2.legend()
            a1.set_xlabel("Time (s)")
            a1.set_ylabel("Wheel Velocity")
            a1.plot(manoeuvre.time, data["v_wheel_fr"], "ROADVIEW3", label="Wheel Velocity FR")
            a1.plot(manoeuvre.time, [i[2] for i in output_states[:].wheel_w_vel], "ROADVIEW3", linestyle ="--", label ="Calculated Wheel Velocity FR")
            a1.set_xlim(xlim)
            a1.legend()
            a3.set_xlabel("Time (s)")
            a3.set_ylabel("Wheel Velocity")
            a3.plot(manoeuvre.time, data["v_wheel_rr"], "ROADVIEW4", label="Wheel Velocity RR")
            a3.plot(manoeuvre.time, [i[3] for i in output_states[:].wheel_w_vel], "ROADVIEW4", linestyle ="--", label ="Calculated Wheel Velocity RR")
            a3.set_xlim(xlim)
            a3.legend()
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/Unoptimized wheel_velocity_subplot-wheelLoad.png")
            # x_rf.fx
            f, ((a0, a1),(a2,a3)) = plt.subplots(2, 2, figsize=figsize, sharex=True,sharey=True)
            
            a0.plot(manoeuvre.time, [i.fy[0] for i in output_states[:].x_rf], "ROADVIEW1", linestyle ="--", label ="Fy FL")
            a0.set_xlim(xlim)
            a0.legend()
            a0.set_xlabel("Time (s)")
            a0.set_ylabel("Wheel Forces in y")
            a2.set_xlabel("Time (s)")
            a2.set_ylabel("Wheel Forces in y")
            a2.plot(manoeuvre.time, [i.fy[1] for i in output_states[:].x_rf], "ROADVIEW2", linestyle ="--", label ="Fy RL")
            a2.set_xlim(xlim)
            a2.legend()
            a1.set_xlabel("Time (s)")
            a1.set_ylabel("Wheel Forces in y")
            a1.plot(manoeuvre.time, [i.fy[2] for i in output_states[:].x_rf], "ROADVIEW3", linestyle ="--", label ="Fy FR")
            a1.set_xlim(xlim)
            a1.legend()
            a3.set_xlabel("Time (s)")
            a3.set_ylabel("Wheel Forces in y")
            a3.plot(manoeuvre.time, [i.fy[3] for i in output_states[:].x_rf], "ROADVIEW4", linestyle ="--", label ="Fy RR")
            a3.set_xlim(xlim)
            a3.legend()
            plt.tight_layout()
            if savefig:
                plt.savefig(fig_save_dir + "/Unoptimized wheel_velocity_subplot-forcesY.png")
 
    plt.show()



from vehicle_dynamics.structures.TireForces import TireForces
from vehicle_dynamics.structures.StateVector import StateVector
from vehicle_dynamics.structures.Displacement import Displacement
from vehicle_dynamics.structures.StrutForce import StrutForce
from vehicle_dynamics.structures.WheelHubForce import WheelHubForce
from vehicle_dynamics.structures.AngularWheelPosition import AngularWheelPosition
from collections import namedtuple
from copy import deepcopy, copy
import numpy as np
OutputState = namedtuple('OutputState', 'compiled_wheel_forces  delta displacement  engine_w f_za f_zr gear powertrain_net_torque slip_x slip_y  sum_f_wheel wheel_w_vel x_a x_rf x_rr suspension_force')


class OutputStates(object):
    """OutputStates Class
        makes a deep copy of the simulation parameters, and stores them in a list, any parameter is available using the .attribute[] operator"""

    def __init__(self):
        super(OutputStates, self).__init__()
        self.compiled_wheel_forces = []
        self.delta = []
        self.displacement = []
        self.engine_w = []
        self.f_za = []
        self.f_zr = []
        self.gear = []
        self.get_data = []
        self.powertrain_net_torque = []
        self.slip_x = []
        self.slip_y = []
        self.sum_f_wheel = []
        self.wheel_w_vel = []
        self.x_a = []
        self.x_rf = []
        self.x_rr = []
        self.suspension_force = []

    def __len__(self):
        return len(self.compiled_wheel_forces)

    def set_states(self, current_states):
        self.compiled_wheel_forces.append(copy(current_states.compiled_wheel_forces))
        self.delta.append(copy(current_states.delta))
        self.displacement.append(copy(current_states.displacement))
        self.engine_w.append(copy(current_states.engine_w))
        self.f_za.append(deepcopy(current_states.f_za))
        self.f_zr.append(deepcopy(current_states.f_zr))
        self.gear.append(copy(current_states.gear))
        self.powertrain_net_torque.append(copy(current_states.powertrain_net_torque))
        self.slip_x.append(copy(current_states.slip_x))
        self.slip_y.append(copy(current_states.slip_y))
        self.sum_f_wheel.append(copy(current_states.sum_f_wheel))
        self.wheel_w_vel.append(copy(current_states.wheel_w_vel))
        self.x_a.append(deepcopy(current_states.x_a))
        self.x_rf.append(deepcopy(current_states.x_rf))
        self.x_rr.append(deepcopy(current_states.x_rr))
        self.suspension_force.append(deepcopy(current_states.suspension_force))

    def __getitem__(self, items):
        return OutputState(self.compiled_wheel_forces,
                           self.delta,
                           self.displacement,
                           self.engine_w,
                           self.f_za,
                           self.f_zr,
                           self.gear,
                           self.powertrain_net_torque,
                           self.slip_x,
                           self.slip_y,
                           self.sum_f_wheel,
                           self.wheel_w_vel,
                           self.x_a,
                           self.x_rf,
                           self.x_rr, 
                           self.suspension_force)

    def padding(self, value):
        self.compiled_wheel_forces.extend([self.compiled_wheel_forces[-1]] * value) 
        self.delta.extend([self.delta[-1]] * value) 
        self.displacement.extend([self.displacement[-1]] * value) 
        self.engine_w.extend([self.engine_w[-1]] * value) 
        self.f_za.extend([self.f_za[-1]] * value) 
        self.f_zr.extend([self.f_zr[-1]] * value) 
        self.gear.extend([self.gear[-1]] * value) 
        self.powertrain_net_torque.extend([self.powertrain_net_torque[-1]] * value) 
        self.slip_x.extend([self.slip_x[-1]] * value) 
        self.slip_y.extend([self.slip_y[-1]] * value) 
        self.sum_f_wheel.extend([self.sum_f_wheel[-1]] * value) 
        self.wheel_w_vel.extend([self.wheel_w_vel[-1]] * value) 
        self.x_a.extend([self.x_a[-1]] * value) 
        self.x_rf.extend([self.x_rf[-1]] * value) 
        self.x_rr.extend([self.x_rr[-1]] * value) 
        self.suspension_force.extend([self.suspension_force[-1]] * value) 
        pass



from setuptools import setup, find_packages

VERSION = '1.0.4' 
DESCRIPTION = 'vehicle_dynamics'
LONG_DESCRIPTION = 'package that calculates the vehicle dynamics'

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="vehicle_dynamics", 
    version=VERSION,
    author="Maikol Funk Dreschler, Yuri Poledna",
    author_email="<Maikol.FunkDreschler@thi.de>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[
        'numpy',
        'matplotlib',
        'scipy',
        "pyyaml",
        "tqdm",
        "pandas",
        "munch",
        "urllib",
    ],

    keywords=['python', 'vehicle dynamics'],
    classifiers= [
        "Programming Language :: Python :: 3",
    ]
)



# -*- coding: utf-8 -*-
"""
Created on Fri Jul  7 11:34:11 2023

@author: albertonbloemer
"""


def import_data_CM(path):
    import pickle 
    with open(path, "rb") as handle:
        data = pickle.load(handle)
        return data 


def main():
    path = "../../exampledata/acc_brake/SimulationData.pickle"
    data = import_data_CM(path)
    print(data)

if __name__ == '__main__':
    main()



import yaml
import numpy as np
from munch import DefaultMunch
from copy import deepcopy
import munch


class StaticParameters(object):
    
    _initialized=False

    def __init__(self, car_parameters_path, freq=1000, optimization=False):
        super(StaticParameters, self).__init__()
        with open(car_parameters_path, 'r') as file:
            param = yaml.safe_load(file)

        self.convert_YAML2Struct(param)

        self.powertrain.gearbox.gear_max_rpm *= np.pi / 30
        self.powertrain.gearbox.gear_min_rpm *= np.pi / 30
        self.powertrain.engine.w_table *= np.pi / 30
        self.powertrain.engine.minimum_rpm *= np.pi / 30
        self.powertrain.engine.maximum_rpm *= np.pi / 30
        self.powertrain.engine.idle_rpm *= np.pi / 30

        self.wheel_base = self.body.lr + self.body.lf
        self.track_width = self.body.wl + self.body.wr

        wd1 = (self.body.lr / (self.wheel_base)) * (self.body.wr / self.track_width)
        wd2 = (self.body.lf / (self.wheel_base)) * (self.body.wr / self.track_width)
        wd3 = (self.body.lr / (self.wheel_base)) * (self.body.wl / self.track_width)
        wd4 = (self.body.lf / (self.wheel_base)) * (self.body.wl / self.track_width)

        self.wd = np.array([wd1, wd2, wd3, wd4])
        self.tire.dynamic_radius = np.array(self.tire.dynamic_radius)
        self.sprung_mass = self.body.mass * self.wd - self.suspension.unsprung_mass

        self.time_step = 1. / freq

        self.aerodynamics_front = self.aerodynamics.air_drag_coefficient * self.aerodynamics.front_area * self.aerodynamics.air_density 

        self.gravity = 9.81

        if not optimization:
            self._initialized = True


    def __setattr__(self, name, value):
        if self._initialized:
            raise AttributeError("We are not in the saving attributes phase of this code, Do Better")
        else:
            object.__setattr__(self, name, value)

    def convert_YAML2Struct(self, yaml):
        data = DefaultMunch.fromDict(yaml)
        for name in dir(data):
            current_key = getattr(data, name)
            setattr(data, name, self.check_munch_key(current_key))
            setattr(self, name, getattr(data, name))

    def check_munch_key(self, current_key):
        for i in dir(current_key):
            if isinstance(getattr(current_key, i), list):
                setattr(current_key, i, np.array(getattr(current_key, i)))
            elif isinstance(getattr(current_key, i), munch.DefaultMunch):
                setattr(current_key, i, self.check_munch_key(getattr(current_key, i)))
        return current_key


def main():
    car_parameters_path = "../../Audi_r8.yaml"

    car_parameters = StaticParameters(car_parameters_path,optimization=True)

    print(car_parameters.aerodynamics.air_drag_coefficient)
    print(car_parameters.powertrain.torque_converter.efficiency_function)
    print(car_parameters.powertrain.torque_converter.efficiency_function)
    car_parameters.gravity = 1
    print(car_parameters.gravity)
    
    car_parameters._initialized = True

    car_parameters.gravity = 9.8
    print(car_parameters.gravity)

if __name__ == '__main__':
    main()
